use crate::parser::template;
use crate::parser::DustTag;
use crate::parser::Template;
use crate::parser::TemplateElement;
use crate::renderer::errors::CompileError;
use crate::renderer::errors::RenderError;
use crate::renderer::renderable::Renderable;
use crate::renderer::walkable::ContextElement;
use crate::renderer::walkable::Walkable;
use std::collections::HashMap;
use std::ops::Index;

#[derive(Clone, Debug)]
pub struct CompiledTemplate<'a> {
    template: Template<'a>,
    pub name: String,
}

#[derive(Clone, Debug)]
pub struct DustRenderer<'a> {
    templates: HashMap<String, &'a Template<'a>>,
}

pub fn compile_template<'a>(
    source: &'a str,
    name: String,
) -> Result<CompiledTemplate<'a>, CompileError> {
    // TODO: Make this all consuming
    // TODO: This could use better error management
    let (_remaining, parsed_template) = template(source).expect("Failed to compile template");
    Ok(CompiledTemplate {
        template: parsed_template,
        name: name,
    })
}

impl<'a> DustRenderer<'a> {
    pub fn new() -> DustRenderer<'a> {
        DustRenderer {
            templates: HashMap::new(),
        }
    }

    pub fn load_source(&mut self, template: &'a CompiledTemplate) {
        self.templates
            .insert(template.name.clone(), &template.template);
    }

    pub fn render<C>(&self, name: &str, context: &C) -> Result<String, RenderError>
    where
        C: ContextElement,
    {
        let main_template = match self.templates.get(name) {
            Some(tmpl) => tmpl,
            None => {
                return Err(RenderError {
                    message: format!("No template named {} in context", name),
                });
            }
        };
        self.render_template(main_template, context)
    }

    fn render_template<C>(&self, template: &Template, context: &C) -> Result<String, RenderError>
    where
        C: ContextElement,
    {
        let mut output = String::new();
        for elem in &template.contents.elements {
            match elem {
                TemplateElement::TESpan(span) => output.push_str(span.contents),
                TemplateElement::TETag(dt) => {
                    output.push_str(&self.render_tag(dt, context)?);
                }
            }
        }
        Ok(output)
    }

    fn render_tag<C>(&self, tag: &DustTag, context: &C) -> Result<String, RenderError>
    where
        C: ContextElement,
    {
        match tag {
            DustTag::DTComment(comment) => (),
            DustTag::DTReference(reference) => {
                let val = walk_path(context, &reference.path.keys);
                return Ok(val.render());
            }
            _ => (), // TODO: Implement the rest
        }
        Ok("".to_owned())
    }
}

fn walk_path<'a>(context: &'a dyn ContextElement, path: &Vec<&str>) -> &'a dyn ContextElement {
    let mut output = context;

    for elem in path.iter() {
        output = output.walk(elem);
    }

    output
}

#[cfg(test)]
mod tests {
    use super::*;

    impl ContextElement for u32 {}
    impl ContextElement for &str {}
    impl<I: ContextElement> ContextElement for HashMap<&str, I> {}

    impl Renderable for u32 {
        fn render(&self) -> std::string::String {
            self.to_string()
        }
    }

    impl Renderable for &str {
        fn render(&self) -> std::string::String {
            self.to_string()
        }
    }

    impl<I: ContextElement> Renderable for HashMap<&str, I> {
        fn render(&self) -> std::string::String {
            panic!("Attempted to render a hashmap");
        }
    }

    impl<I: ContextElement> Walkable for HashMap<&str, I> {
        fn walk(&self, segment: &str) -> &dyn ContextElement {
            self.get(segment).unwrap()
        }
    }

    impl Walkable for &str {
        fn walk(&self, _segment: &str) -> &dyn ContextElement {
            panic!("Tried to walk down a str");
        }
    }

    impl Walkable for u32 {
        fn walk(&self, _segment: &str) -> &dyn ContextElement {
            panic!("Tried to walk down a str");
        }
    }

    #[test]
    fn test_walk_path() {
        let context: HashMap<&str, &str> =
            [("cat", "kitty"), ("dog", "doggy"), ("tiger", "murderkitty")]
                .iter()
                .cloned()
                .collect();
        let number_context: HashMap<&str, u32> = [("cat", 1), ("dog", 2), ("tiger", 3)]
            .iter()
            .cloned()
            .collect();
        let deep_context: HashMap<&str, HashMap<&str, &str>> = [
            ("cat", [("food", "meat")].iter().cloned().collect()),
            ("dog", [("food", "meat")].iter().cloned().collect()),
            ("tiger", [("food", "people")].iter().cloned().collect()),
        ]
        .iter()
        .cloned()
        .collect();
        assert_eq!(walk_path(&context, &vec!["cat"]).render(), "kitty");
        assert_eq!(walk_path(&number_context, &vec!["tiger"]).render(), "3");
        assert_eq!(
            walk_path(&deep_context, &vec!["tiger", "food"]).render(),
            "people"
        );
    }
}
